﻿//License: GPLv3 (see LICENSE.txt)

using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System;
using UnityEngine;
using Toolbar;

namespace Rotate
{
    [KSPAddon(KSPAddon.Startup.Flight, false)]
    public class VehicleRotator : MonoBehaviour
    {
        static bool started = false;
        static bool loaded = false;
        static public Dictionary<string, Vector3> vessel_momentum = new Dictionary<string, Vector3>();
        static public Dictionary<string, Quaternion> vessel_rotation = new Dictionary<string, Quaternion>();
        static public Dictionary<string, Vector3> vessel_direction = new Dictionary<string, Vector3>();
        static public Dictionary<string, CelestialBody> vessel_reference = new Dictionary<string, CelestialBody>();

        static public Dictionary<ProtoCrewMember, float> ArtificialGravity = new Dictionary<ProtoCrewMember, float>();

        static public Dictionary<string, Vector3> lastPosition = new Dictionary<string, Vector3>();
        static public Dictionary<string, Transform> lastTransform = new Dictionary<string, Transform>();

        private static String cfg
        {
            get { return KSPUtil.ApplicationRootPath + "/GameData/PersistentRotation/PersistentRotation_" + HighLogic.CurrentGame.Title.TrimEnd( "_()SANDBOXCAREERSCIENCE".ToCharArray() ).TrimEnd(' ') + ".cfg"; }
        }
        int ticker = 0;
        protected Rect windowPos;
        bool show = true;
        bool unit = false;
        IButton Tbutton;
        Vessel lastVessel;

        public void WindowGUI(int windowID)
        {
            GUIStyle mySty = new GUIStyle(GUI.skin.button);
            mySty.normal.textColor = mySty.focused.textColor = Color.white;
            mySty.hover.textColor = mySty.active.textColor = Color.yellow;
            mySty.onNormal.textColor = mySty.onFocused.textColor = mySty.onHover.textColor = mySty.onActive.textColor = Color.green;
            mySty.padding = new RectOffset(8, 8, 8, 8);

            GUILayout.BeginVertical();
            if (GUILayout.Button("change units", mySty, GUILayout.ExpandWidth(true)))//GUILayout.Button is "true" when clicked
            {
                unit = !unit;
            }

            foreach (ProtoCrewMember kerb in FlightGlobals.ActiveVessel.GetVesselCrew())
            {
                if (FlightGlobals.ActiveVessel.vesselType != VesselType.EVA
                || FlightGlobals.ActiveVessel.vesselType != VesselType.Flag
                || FlightGlobals.ActiveVessel.vesselType != VesselType.Unknown)
                {
                    if (kerb.KerbalRef.state != Kerbal.States.DEAD)
                    {
                        GUILayout.BeginHorizontal();
                        GUILayout.Label(kerb.name.TrimEnd("Kerman".ToCharArray()));
                        GUILayout.FlexibleSpace();
                        if (unit)
                        {
                            CalculateArtificialGeForce();
                            ArtificialGravity[kerb] = (float)((int)(ArtificialGravity[kerb] * 10)) / 10;
                            GUILayout.Label(ArtificialGravity[kerb].ToString() + " m/s");
                        }
                        else
                        {
                            CalculateArtificialGeForce();
                            ArtificialGravity[kerb] = (float)((int)((ArtificialGravity[kerb] / 9.81) * 10)) / 10;
                            GUILayout.Label(ArtificialGravity[kerb].ToString() + " G");
                        }
                        GUILayout.EndHorizontal();
                    }
                }
            }

            GUILayout.EndVertical();
            GUI.DragWindow(new Rect(0, 0, 10000, 20));

        }
        private void drawGUI()
        {
            GUI.skin = null;
            GUISkin unitySkin = (GUISkin)GameObject.Instantiate(GUI.skin);
            GUI.skin = unitySkin;

            windowPos = GUILayout.Window(1, windowPos, WindowGUI, "Artificial Gravity", GUILayout.MinWidth(10));
        }

        public void CleanCfg()
        {
            //Clean vessel momentum
            Dictionary<string, Vector3> temp_vessel_momentum = new Dictionary<string, Vector3>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_momentum[vessel.id.ToString()] = Vector3.zero;
            }
            foreach (KeyValuePair<string, Vector3> entry in vessel_momentum)
            {
                temp_vessel_momentum[entry.Key] = entry.Value;
            }
            vessel_momentum.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_momentum[vessel.id.ToString()] = temp_vessel_momentum[vessel.id.ToString()];
            }
            temp_vessel_momentum.Clear();
            //Clean vessel rotation
            Dictionary<string, Quaternion> temp_vessel_rotation = new Dictionary<string, Quaternion>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_rotation[vessel.id.ToString()] = Quaternion.identity;
            }
            foreach (KeyValuePair<string, Quaternion> entry in vessel_rotation)
            {
                temp_vessel_rotation[entry.Key] = entry.Value;
            }
            vessel_rotation.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_rotation[vessel.id.ToString()] = temp_vessel_rotation[vessel.id.ToString()];
            }
            temp_vessel_rotation.Clear();
            //Clean vessel direction
            Dictionary<string, Vector3> temp_vessel_direction = new Dictionary<string, Vector3>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_direction[vessel.id.ToString()] = Vector3.zero;
            }
            foreach (KeyValuePair<string, Vector3> entry in vessel_direction)
            {
                temp_vessel_direction[entry.Key] = entry.Value;
            }
            vessel_direction.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_direction[vessel.id.ToString()] = temp_vessel_direction[vessel.id.ToString()];
            }
            temp_vessel_direction.Clear();
            //Clean vessel reference
            Dictionary<string, CelestialBody> temp_vessel_reference = new Dictionary<string, CelestialBody>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_reference[vessel.id.ToString()] = null;
            }
            foreach (KeyValuePair<string, CelestialBody> entry in vessel_reference)
            {
                temp_vessel_reference[entry.Key] = entry.Value;
            }
            vessel_reference.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_reference[vessel.id.ToString()] = temp_vessel_reference[vessel.id.ToString()];
            }
            temp_vessel_reference.Clear();
        }
        public void Save()
        {
            try
            {
                CleanCfg();

                ConfigNode save = new ConfigNode();
                ConfigNode.CreateConfigFromObject(this, 0, save);

                //Save momentum of all vessels
                ConfigNode momentum = save.AddNode("MOMENTUM");
                foreach (KeyValuePair<string, Vector3> entry in vessel_momentum)
                {
                    momentum.AddValue(entry.Key, KSPUtil.WriteVector(entry.Value));
                }
                //Save rotation of all vessels
                ConfigNode rotation = save.AddNode("ROTATION");
                foreach (KeyValuePair<string, Quaternion> entry in vessel_rotation)
                {
                    rotation.AddValue(entry.Key, KSPUtil.WriteQuaternion(entry.Value));
                }
                //Save direction to the mainBody for each vessel
                ConfigNode direction = save.AddNode("DIRECTION");
                foreach (KeyValuePair<string, Vector3> entry in vessel_direction)
                {
                    direction.AddValue(entry.Key, KSPUtil.WriteVector(entry.Value));
                }
                //Save direction to the mainBody for each vessel
                ConfigNode reference = save.AddNode("REFERENCE");
                foreach (KeyValuePair<string, CelestialBody> entry in vessel_reference)
                {
                    if (entry.Value != null)
                    {
                        reference.AddValue(entry.Key, entry.Value.name);
                    }
                    else
                    {
                        reference.AddValue(entry.Key, "NONE");
                    }
                }
                save.Save(cfg);
            }
            catch (Exception e) { print("[PR] Saving not sucessfull: " + e.Message); }
        }
        public static void Load()
        {
            //This is called when all persistent rotation data is being loaded from the cfg file.

            loaded = true;
            ConfigNode load = ConfigNode.Load(cfg);
            if (load == null)
            {
                print("[PR] Save file is empty or not existent!");
                return;
            }
            //vessel_momentum from scratch
            vessel_momentum.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_momentum[vessel.id.ToString()] = Vector3.zero;
            }
            ConfigNode momentum = load.GetNode("MOMENTUM");
            foreach (ConfigNode.Value s in momentum.values)
            {
                vessel_momentum[s.name] = KSPUtil.ParseVector3(s.value);
            }
            //vessel_rotation from scratch
            vessel_rotation.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_rotation[vessel.id.ToString()] = vessel.transform.rotation;
            }
            ConfigNode rotation = load.GetNode("ROTATION");
            foreach (ConfigNode.Value s in rotation.values)
            {
                vessel_rotation[s.name] = KSPUtil.ParseQuaternion(s.value);
            }

            //vessel_direction from scratch
            vessel_direction.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_direction[vessel.id.ToString()] = vessel.mainBody.position - vessel.transform.position;
            }
            ConfigNode direction = load.GetNode("DIRECTION");
            foreach (ConfigNode.Value s in direction.values)
            {
                vessel_direction[s.name] = KSPUtil.ParseVector3(s.value);
            }
            //vessel_reference from scratch
            vessel_reference.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_reference[vessel.id.ToString()] = null;
            }
            ConfigNode reference = load.GetNode("REFERENCE");
            foreach (ConfigNode.Value s in reference.values)
            {
                if (s.value != "NONE")
                {
                    foreach (CelestialBody body in FlightGlobals.Bodies)
                    {
                            if (body.name == s.value)
                            {
                                vessel_reference[s.name] = body;
                            }
                    }
                }
                else
                {
                    vessel_reference[s.name] = null;
                }
            }
        }

        void OnVesselGoOnRails(Vessel vessel)
        {
            if (vessel.ActionGroups[KSPActionGroup.SAS])
            {
                //If Autopilot is enabled when vessel goes on rails
                vessel_momentum[vessel.id.ToString()] = Vector3.zero;
            }
            else
            {
                vessel_momentum[vessel.id.ToString()] = Quaternion.Inverse( Quaternion.Inverse( lastVessel.GetReferenceTransformPart().transform.rotation) * lastVessel.transform.rotation) * vessel.angularVelocity; //HERE
            }

            vessel_rotation[vessel.id.ToString()] =  vessel.transform.rotation;

            if (vessel_reference[vessel.id.ToString()] != null)
            {
                vessel_direction[vessel.id.ToString()] = vessel_reference[vessel.id.ToString()].position - vessel.transform.position;
            }
            Save();
        }
        void OnVesselGoOffRails(Vessel vessel)
        {
            Load();
            CleanCfg();

            if (vessel_reference[vessel.id.ToString()] != null && vessel.ActionGroups[KSPActionGroup.SAS])
            {
                vessel.SetRotation(Quaternion.FromToRotation(vessel_direction[vessel.id.ToString()], vessel_reference[vessel.id.ToString()].position - vessel.transform.position) * vessel_rotation[vessel.id.ToString()]);
            }
            else
            {
                try
                {

                    Vector3 av = vessel_momentum[vessel.id.ToString()];
                    Vector3 COM = vessel.findWorldCenterOfMass();
                    foreach (Part p in vessel.parts) //Applying force on every part
                    {
                        try
                        {
                            if (p.rigidbody == null) continue;
                            p.rigidbody.AddTorque(vessel.transform.rotation * av, ForceMode.VelocityChange);
                            p.rigidbody.AddForce(Vector3.Cross(vessel.transform.rotation * av, (p.rigidbody.position - COM)), ForceMode.VelocityChange);
                        }
                        catch (NullReferenceException nre)
                        {
                            print("[PR] NullReferenceException: " + nre.Message);
                        }

                    }
                }
                catch (Exception e)
                {
                    print("[PR] Exception: " + e.Message);
                    // Don't really need to do anything; we just haven't met this vehicle yet.
                }
            }

        }

        void Start()
        {
            if (!started)
            {
                GameEvents.onVesselGoOnRails.Add(OnVesselGoOnRails);
                GameEvents.onVesselGoOffRails.Add(OnVesselGoOffRails);
                GameEvents.onGamePause.Add(OnPause);
                started = true;
            }

            CleanCfg();

            if ((windowPos.x == 0) && (windowPos.y == 0))
            {
                windowPos = new Rect(Screen.width / 4, 0, 150, 10);
            }

            foreach(Vessel vessel in FlightGlobals.Vessels)
            {
                lastTransform[vessel.id.ToString()] = vessel.ReferenceTransform;
                if (vessel_reference[vessel.id.ToString()] != null)
                {
                    lastPosition[vessel.id.ToString()] = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].position;
                }
            }
        }
        void OnPause()
        {
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                if (!vessel.packed)
                {
                    vessel_momentum[vessel.id.ToString()] = vessel.angularVelocity;
                    vessel_rotation[vessel.id.ToString()] = vessel.transform.rotation;
                    if (vessel_reference[vessel.id.ToString()] != null)
                    {
                        vessel_direction[vessel.id.ToString()] = vessel_reference[vessel.id.ToString()].position - vessel.transform.position;
                    }
                }
            }
            Save();
        }

        public static QuaternionD FromToRotation(Vector3d fromv, Vector3d tov)
        {
            // Doesn't handle the singularity with precisely opposite vectors
            Vector3d cross = Vector3d.Cross(fromv, tov);
            double dot = Vector3d.Dot(fromv, tov);
            double wval = dot + Math.Sqrt(fromv.sqrMagnitude * tov.sqrMagnitude);
            double norm = 1.0 / Math.Sqrt(cross.sqrMagnitude + wval * wval);
            return new QuaternionD(cross.x * norm, cross.y * norm, cross.z * norm, wval * norm);
        }

        public void SurfaceSAS(Vessel vessel)
        {
            if (vessel_reference[vessel.id.ToString()] != null) //Check vessel_reference in current frame
            {
                Vector3d newPosition = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].position;
                QuaternionD delta = FromToRotation(lastPosition[vessel.id.ToString()], newPosition);
                QuaternionD adjusted = delta * (QuaternionD)vessel.Autopilot.SAS.lockedHeading;
                vessel.Autopilot.SAS.lockedHeading = adjusted;

                //TODO: optimize for Active Vessel
            }
        }
        public void RotateWhilePacked(Vessel vessel)
        {
            Vector3 angvel = vessel_momentum[vessel.id.ToString()];

            if (vessel.situation != Vessel.Situations.LANDED
                && vessel.situation != Vessel.Situations.PRELAUNCH
                && vessel.situation != Vessel.Situations.DOCKED
                && vessel.situation != Vessel.Situations.SPLASHED)
            {
                //Either the vessel continues to rotate, or it follows the planets curvature. This might change.
                if (vessel.ActionGroups[KSPActionGroup.SAS])
                {
                    //TODO: Continue to rotate
                    if (vessel_reference[vessel.id.ToString()] != null)
                    {
                        vessel.SetRotation(Quaternion.FromToRotation(vessel_direction[vessel.id.ToString()], vessel_reference[vessel.id.ToString()].position - vessel.transform.position) * vessel_rotation[vessel.id.ToString()]);
                    }
                }
                else
                {
                    float angle = angvel.magnitude * TimeWarp.CurrentRate;
                    Quaternion rotation = Quaternion.AngleAxis(angle, angvel);
                    vessel.SetRotation(vessel.transform.rotation * rotation);

                    if(vessel == FlightGlobals.ActiveVessel)
                    {
                        print("angvel: " + angvel.ToString());
                        print("rotation: " + rotation.ToString());
                        print("vessel.transform.rotation: " + vessel.transform.rotation.ToString());
                    }
                }
            }
        }
        public void CalculateArtificialGeForce()
        {
            if (FlightGlobals.ActiveVessel.GetCrewCount() > 0)
            {
                foreach (ProtoCrewMember kerb in FlightGlobals.ActiveVessel.GetVesselCrew())
                {
                    if (kerb.seat)
                    {
                        float speed = FlightGlobals.ActiveVessel.angularVelocity.magnitude * (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;
                        float force = (speed * speed) / (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;
                        ArtificialGravity[kerb] = force;
                        if (ArtificialGravity[kerb] > 10 * 9.81)
                        {
                            ScreenMessages.PostScreenMessage("Danger: Crew Member reaching Artificial Gravity Limit!");
                        }
                        if (force > 9.81 * 22 - 9.81 * kerb.KerbalRef.stupidity * 4)
                        {
                            kerb.KerbalRef.die(); //Let Kerbal die if GeForce exceeds ~20 G's
                            ArtificialGravity.Remove(kerb);
                            windowPos.height = 10;
                        }
                    }
                }
            }
        }
        void FixedUpdate()
        {
            if (FlightGlobals.ActiveVessel != lastVessel)
                CleanCfg();

            ticker++;
            if (ticker > 100)
            {
                ticker = 0;
                CleanCfg();
            }
            if (!loaded)
            {
                Load();
            }
            try
            {
                foreach (Vessel vessel in FlightGlobals.Vessels)
                {
                    if (vessel.packed)
                    {
                        RotateWhilePacked(vessel);
                    }
                    else
                    {
                        SurfaceSAS(vessel);
                    }
                    lastTransform[vessel.id.ToString()] = vessel.ReferenceTransform;
                    if (vessel_reference[vessel.id.ToString()] != null)
                    {
                        lastPosition[vessel.id.ToString()] = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].position;
                    }
                }
            }
            catch (NullReferenceException nre)
            {
                print("NullReference: " + nre.ToString());
            }

            lastVessel = FlightGlobals.ActiveVessel;

            print( Quaternion.Inverse(lastVessel.GetReferenceTransformPart().transform.rotation) * lastVessel.transform.rotation );
        }

        internal VehicleRotator()
        {
            Tbutton = ToolbarManager.Instance.add("persistentrotation", "button");
            Tbutton.TexturePath = "PersistentRotation/Textures/texture";
            Tbutton.ToolTip = "Persistent Rotation";
            Tbutton.Visibility = new GameScenesVisibility(GameScenes.FLIGHT);
            Tbutton.OnClick += (e) => togglePopupMenu(Tbutton);
        }
        private void togglePopupMenu(IButton button)
        {
            if (button.Drawable == null && FlightGlobals.ActiveVessel.packed == false)
            {
                createPopupMenu(button);
            }
            else
            {
                destroyPopupMenu(button);
            }
        }
        private void createPopupMenu(IButton button)
        {
            // create menu drawable
            PopupMenuDrawable menu = new PopupMenuDrawable();
            // create menu options
            IButton GUItoggle = menu.AddOption("toggle GUI");
            GUItoggle.OnClick += (e2) =>
            {
                if (show == true)
                {
                    RenderingManager.AddToPostDrawQueue(3, new Callback(drawGUI));
                }
                else
                {
                    RenderingManager.RemoveFromPostDrawQueue(3, new Callback(drawGUI));
                }
                show = !show;
            };
            menu.AddSeparator();
            IButton SASclear = menu.AddOption("SAS clear");
            SASclear.OnClick += (e2) => { CleanCfg(); vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = null; };
            menu.AddSeparator();
            foreach (CelestialBody body in FlightGlobals.Bodies)
            {
                IButton bodylock = menu.AddOption(body.name.ToString());
                bodylock.OnClick += (e2) => { CleanCfg(); vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = body; };
            }
            // auto-close popup menu when any option is clicked
            menu.OnAnyOptionClicked += () => destroyPopupMenu(button);
            // hook drawable to button
            button.Drawable = menu;
        }
        private void destroyPopupMenu(IButton button)
        {
            // PopupMenuDrawable must be destroyed explicitly
            ((PopupMenuDrawable)button.Drawable).Destroy();
            // unhook drawable
            button.Drawable = null;
        }
        internal void OnDestroy()
        {
            Tbutton.Destroy();
        }
    }
}